/*
Copyright (c) 2005 Aristotle Pagaltzis

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#define _GNU_SOURCE
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <err.h>

#include <libxml/tree.h>
#include <libxml/parser.h>
#include <libxml/xpath.h>
#include <libxml/xpathInternals.h>

#if !(defined(LIBXML_XPATH_ENABLED) && defined(LIBXML_SAX1_ENABLED))
#	error "Your copy of libxml2 is not compiled with support for the necessary APIs (XPath and SAX1)"
#endif

int usage_error = 0;

xmlDocPtr doc = NULL, newdoc = NULL;
xmlXPathContextPtr xpathctx = NULL;
xmlXPathObjectPtr xpathobj = NULL;
xmlHashTablePtr ns_registry = NULL;
xmlListPtr expr_list = NULL;

void
cleanup( void )
{
	if( usage_error ) fprintf( stderr,
		"usage: xpathgrep [-w pfx:tag] [-n pfx:uri] [-e xpath] file.xml [...]\n"
		"        -w  Name of document element to wrap all results; if you do not\n"
		"            use a prefix, the element will be in the default namespace.\n"
		"            If a wrapper is not requested, the results will be given\n"
		"            individually.\n"
		"        -n  Namespace to register for use in XPath; syntax is prefix,\n"
		"            followed by the namespace URI, separated by a semicolon,\n"
		"            eg. ``html:http://www.w3.org/1999/xhtml'' to bind the URI\n"
		"            ``http://www.w3.org/1999/xhtml'' to the prefix ``html''. If\n"
		"            you have requested a wrapper, all registered namespaces\n"
		"            will also be declared on it. To declare a default\n"
		"            namespace on the wrapper, register a namespace with an\n"
		"            empty prefix, eg ``-e :http://www.w3.org/1999/xhtml''.\n"
		"        -e  An XPath expression; will be evaluated against each\n"
		"            document. For a single expression you can omit the switch\n"
		"            and just give it before any filenames.\n"
		"        -h  This help.\n"
		"\n"
		"        You can use a filename of ``-'' to denote standard input.\n"
		"\n"
		"        Note that your output will probably have lots of redundant\n"
		"        namespace declarations; libxml2 offers no easy way to clean\n"
		"        them up at the time of writing. If you want to clean up a result\n"
		"        document, pipe it to ``xmllint --nsclean -''.\n"
	);
	if( xpathobj ) xmlXPathFreeObject( xpathobj );
	if( xpathctx ) xmlXPathFreeContext( xpathctx );
	if( doc ) xmlFreeDoc( doc );
	if( newdoc ) xmlFreeDoc( newdoc );
	if( ns_registry ) xmlHashFree( ns_registry, NULL );
	if( expr_list ) xmlListDelete( expr_list );
	xmlCleanupParser();
}

void
print_xmldoc( xmlDocPtr doc )
{
	xmlBufferPtr buf = xmlBufferCreate();
	xmlNodeDump( buf, doc, xmlDocGetRootElement( doc ), 0, 0 );
	printf( "%s\n", ( char * ) xmlBufferContent( buf ) );
	xmlBufferFree( buf );
}

void
nsdecl_parse_register( xmlHashTablePtr ns_registry, char * decl )
{
	char * prefix = decl;
	char * uri = strchr( prefix, ':' );
	if( uri ) { *uri = 0; ++uri; } else { uri = ""; }
	xmlHashAddEntry( ns_registry, xmlCharStrdup( prefix ), xmlCharStrdup( uri ) );
}

void
node_declare_namespace( xmlChar * uri, xmlNodePtr node, xmlChar * prefix )
{
	if( ! *prefix ) prefix = NULL;
	xmlNsPtr ns = xmlNewNs( node, uri, prefix );
	if( ! ns ) errx( -1, "unable to declare NS with prefix \"%s\" bound to URI \"%s\"", prefix, uri );
}

void
xpath_register_namespace( xmlChar * uri, xmlXPathContextPtr xpathctx, xmlChar * prefix )
{
	int failure = xmlXPathRegisterNs( xpathctx, prefix, uri );
	if( failure ) errx( -1, "unable to register NS with prefix \"%s\" bound to URI \"%s\"", prefix, uri );
}

void
setup_wrapperdoc( char * doc_element )
{
	xmlNsPtr ns = NULL;
	xmlNodePtr doc_elt = NULL;
	char * prefix = doc_element;
	char * elt_name = strchr( doc_element, ':' );

	if( elt_name ) { *elt_name = 0; ++elt_name; }
	else { prefix = NULL; elt_name = doc_element; }

	doc_elt = xmlNewDocNode( newdoc, NULL, BAD_CAST elt_name, BAD_CAST "\n" );
	xmlDocSetRootElement( newdoc, doc_elt );

	xmlHashScan( ns_registry, (xmlHashScanner) node_declare_namespace, doc_elt );

	ns = xmlSearchNs( newdoc, doc_elt, BAD_CAST prefix );
	if( ns ) xmlSetNs( doc_elt, ns );
}

int
eval_xpath( const xmlChar * xpath, const xmlDocPtr doc )
{
	int size, i;

	xpathctx = xmlXPathNewContext( doc );
	if( !xpathctx ) errx( -1, "unable to create new XPath context" );

	xmlHashScan( ns_registry, (xmlHashScanner) xpath_register_namespace, xpathctx );

	xpathobj = xmlXPathEvalExpression( xpath, xpathctx );
	if( !xpathobj ) errx( 3, "unable to evaluate xpath expression: %s", xpath );

	size = ( xpathobj->nodesetval ) ? xpathobj->nodesetval->nodeNr : 0;

	/* Reserialise fragment by copying it into a new document */
	for( i = 0; i < size; ++i ) {
		xmlNodePtr root = xmlDocGetRootElement( newdoc );
		xmlNodePtr copied = xmlDocCopyNode( xpathobj->nodesetval->nodeTab[i], newdoc, 1 );

		if( root ) {
			xmlAddChild( root, copied );
		}
		else {
			xmlDocSetRootElement( newdoc, copied );
			print_xmldoc( newdoc );
			/* xmlFreeDoc would clean the node up anyway, but in order to be parsimonious... */
			xmlUnlinkNode( copied );
			xmlFreeNode( copied );
		}
	}

	if( xpathobj ) xmlXPathFreeObject( xpathobj ), xpathobj = NULL;
	if( xpathctx ) xmlXPathFreeContext( xpathctx ), xpathctx = NULL;

	return 1;
}

int
main( int argc, char **argv )
{
	int i = 0;
	char * wrapper = NULL;

	atexit( cleanup );

	ns_registry = xmlHashCreate( 0 );

	expr_list = xmlListCreate( NULL, NULL );

	if( argc <= 1 ) ++usage_error, exit( 0 );

	while( ( i = getopt( argc, argv, "e:n:w:h" ) ) != -1 ) {
		switch( i ) {
			case 'e':
				xmlListAppend( expr_list, xmlCharStrdup( optarg ) );
				break;
			case 'n':
				nsdecl_parse_register( ns_registry, optarg );
				break;
			case 'w':
				wrapper = strdup( optarg );
				if( ! wrapper ) errx( -1, "out of memory" );
				break;
			case 'h':
				++usage_error, exit( 0 );
			default:
				++usage_error, exit( 1 );
		}
	}

	if( ( optind < argc ) && ( xmlListSize( expr_list ) == 0 ) )
		xmlListAppend( expr_list, xmlCharStrdup( argv[ optind++ ] ) );

	if( optind >= argc )
		++usage_error, errx( 1, "no file to search" );

	if( xmlListSize( expr_list ) == 0 )
		++usage_error, errx( 1, "no XPath expression given" );

	xmlInitParser();
	LIBXML_TEST_VERSION;

	newdoc = xmlNewDoc( BAD_CAST "1.0" );
	if( !newdoc ) errx( -1, "unable to allocate result document" );

	if( wrapper ) setup_wrapperdoc( wrapper );

	for( ; optind < argc ; ++optind ) {
		char * filename = argv[ optind ];
		doc = xmlParseFile( filename );
		if( !doc ) errx( 2, "unable to parse file: %s", filename );
		xmlListWalk( expr_list, (xmlListWalker) eval_xpath, doc );
		xmlFreeDoc( doc ), doc = NULL;
	}

	if( wrapper ) {
		print_xmldoc( newdoc );
		free( wrapper ), wrapper = NULL;
	}

	return 0;
}
